/*
 * LiPa module Huecycler - cycling hue (H of HSV) of each pixel
 * v0.5.0, 2018.02.12
 * Copyright (c) 2018 Sunkware
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * WWW: sunkware.org				E-MAIL: sunkware@gmail.com
 */


// "Namespace"

LiPa.Huecycler = {};
LiPa.Huecycler.modname = "Huecycler";

// Primary parameters

LiPa.Huecycler.duration = 8000;
LiPa.Huecycler.frate = 30;

// State

LiPa.Huecycler.ctrlEnt = null;

LiPa.Huecycler.beginTime = 0;

LiPa.Huecycler.timer = -1;

LiPa.Huecycler.secondRatioFromHueMod2 = null;
LiPa.Huecycler.chromaComponentIndexFromHueSector = null;
LiPa.Huecycler.secondComponentIndexFromHueSector = null;

LiPa.Huecycler.ScrH = null;
LiPa.Huecycler.ScrS = null;
LiPa.Huecycler.ScrV = null;

// Functions

LiPa.Huecycler.initColorTables = function() {
// Table of "1 - |(hue mod 2) - 1|"
	LiPa.Huecycler.secondRatioFromHueMod2 = new Array(0x200);
	for (var i = 0; i < 0x200; i++) {
		LiPa.Huecycler.secondRatioFromHueMod2[i] = 0x100 - Math.abs(i - 0x100);
	}

// Position of component in RGB color
	LiPa.Huecycler.chromaComponentIndexFromHueSector = new Array(0, 1, 1, 2, 2, 0);
	LiPa.Huecycler.secondComponentIndexFromHueSector = new Array(1, 0, 2, 1, 0, 2);
}

// Color space conversion

LiPa.Huecycler.RGB2HSV = function(red, green, blue) { // r, g, b are in [0;1]
	var cmax = Math.max(Math.max(red, green), blue);
	var cmin = Math.min(Math.min(red, green), blue);
	var chroma = cmax - cmin;

	var value = cmax;
	var saturation = cmax > 0.0 ? chroma / cmax : 0;
	var hue;

	if (chroma == 0.0) {
		hue = 0.0;
	} else if (cmax == red) {
		hue = (6.0 + (green - blue) / chroma) % 6.0;		
	} else if (cmax == green) {
		hue = 2 + (blue - red) / chroma;
	} else { // cmax == blue
		hue = 4 + (red - green) / chroma;
	}

// hue *= 60;
// Now: H in [0;6], S in [0;1], V in [0;1]

	return [Math.floor(hue * 0x100), Math.floor(saturation * 0x400), Math.floor(value * 0x100)];
}

LiPa.Huecycler.HSV2RGB = function(hue, saturation, value) { // H in [0;6] * 0x100, S in [0;1] * 0x400, V in [0;1] * 0x100
// Need to be performance-optimized, since it's called for each pixel of canvas
	var hueSector = hue >> 8;

	var chroma = (saturation * value) >> 0xA;

	var second = (chroma * LiPa.Huecycler.secondRatioFromHueMod2[hue & 0x1FF]) >> 8;

	var cmin = value - chroma;

	var chromaIndex = LiPa.Huecycler.chromaComponentIndexFromHueSector[hueSector];
	var secondIndex = LiPa.Huecycler.secondComponentIndexFromHueSector[hueSector];

	return ( ((chroma + cmin) << (chromaIndex << 3)) + ((second + cmin) << (secondIndex << 3)) + (cmin << ((3 - chromaIndex - secondIndex) << 3)) + 0xFF000000 ); // alpha = 0xFF
}

LiPa.Huecycler.convertScrToHSV = function() {
	var scrData = LiPa.Huecycler.ctrlEnt.scrData;
	var width = LiPa.Huecycler.ctrlEnt.width;
	var height = LiPa.Huecycler.ctrlEnt.height;

	LiPa.Huecycler.ScrH = new Array(width * height);
	LiPa.Huecycler.ScrS = new Array(width * height);
	LiPa.Huecycler.ScrV = new Array(width * height);

	var offs = 0, soffs = 0;
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			hsv = LiPa.Huecycler.RGB2HSV(scrData.data[soffs] / 0x100, scrData.data[soffs + 1] / 0x100, scrData.data[soffs + 2] / 0x100);
			LiPa.Huecycler.ScrH[offs] = hsv[0];
			LiPa.Huecycler.ScrS[offs] = hsv[1];
			LiPa.Huecycler.ScrV[offs] = hsv[2];

			offs++;
			soffs += 4;
		}
	}
}

LiPa.Huecycler.drawHueshifted = function(shift) {
	var drawData = LiPa.Huecycler.ctrlEnt.drawData;
	var width = LiPa.Huecycler.ctrlEnt.width;
	var height = LiPa.Huecycler.ctrlEnt.height;

	shift = Math.floor(shift * 0x600);
	var hueMapping = new Array(0x600);
	for (i = 0; i < 0x600; i++) {
		hueMapping[i] = (i + shift) % 0x600;
	}

	var offs = 0, doffs = 0;

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			rgb = LiPa.Huecycler.HSV2RGB(hueMapping[LiPa.Huecycler.ScrH[offs]], LiPa.Huecycler.ScrS[offs], LiPa.Huecycler.ScrV[offs]);

			drawData.data[doffs] = rgb & 0xFF;
			drawData.data[doffs + 1] = (rgb >> 0x8) & 0xFF;
			drawData.data[doffs + 2] = (rgb >> 0x10) & 0xFF;
			drawData.data[doffs + 3] = 0xFF;

			offs++;
			doffs += 4;
		}
	}
}

LiPa.Huecycler.main = function() {
	var shift = ((Date.now() - LiPa.Huecycler.beginTime) % LiPa.Huecycler.duration) / LiPa.Huecycler.duration;

	LiPa.Huecycler.drawHueshifted(shift);

	LiPa.showDirect();

	LiPa.Huecycler.timer = window.setTimeout(LiPa.Huecycler.main, Math.floor(1000 / LiPa.Huecycler.frate));
}

LiPa.Huecycler.start = function(ctrlEnt) {
	console.log(LiPa.Huecycler.modname + " started");

	LiPa.Huecycler.ctrlEnt = ctrlEnt;
	LiPa.Huecycler.initColorTables();
	LiPa.Huecycler.convertScrToHSV();
	LiPa.Huecycler.beginTime = Date.now();
	LiPa.Huecycler.main();	
}

LiPa.Huecycler.stop = function() {
	if (LiPa.Huecycler.timer != -1) {
		window.clearTimeout(LiPa.Huecycler.timer);
		LiPa.Huecycler.timer = -1;
	}

	console.log(LiPa.Huecycler.modname + " stopped");
}

// Registering

LiPa.addStartStopFunctions(LiPa.Huecycler.start, LiPa.Huecycler.stop);
